unit EventHandlers;

interface
uses
  Generics.Collections, System.Classes, System.SysUtils, IObj;

type
  {
    // usage
    type
      TMyClass = class(...)
      private
        F<eventName> : TEventHandlers<EventType>;
      public
        property <eventName> : TEventHandlers<EventType> read F<EventName> write Set<EventName>
      end;

    ...
    var myObj := TMyClass;
    ...

    // support the following syntaxes
    myObj.<eventName>.setHandler(someEventHandler);
    myObj.<eventName>.clear;
    myObj.<eventName>.execute(...);

    myObj.<eventName>.addHandler(someEventHandler);
    myObj.<eventName>.removeHandler(someEventHandler);

    myObj.<eventName>.subscribe(anObject, someEventHandler);
    myObj.<eventName>.unsubscribe(anObject);
  }

  // T is the eventHandler type
  // eg. TNotifyEvent, TProc, TProc<T>, TFunc<T>
  TEventHandlers<T> = class(TComponent)
  type
    TEventHandler = class
      subscriber: TObject;
      eventHandler: T;
    public
      constructor Create(obj: TObject; handler: T);
    end;
    IEventHandler = IObj<TEventHandler>;
    TIEventHandlerList = TList<IEventHandler>;
    IIEventHandlerList = IObj<TIEventHandlerList>;

  private
    fList: IIEventHandlerList;
    procedure createList;

  protected
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
  public
    procedure setHandler(handler: T);
    procedure addHandler(handler: T);
    procedure removeHandler(handler: T);

    procedure subscribe(subscriber: TObject; handler: T);
    procedure unsubscribe(subscriber: TObject); overload;
    procedure unsubscribe(subscriber: TObject; handler: T); overload;

    procedure clear;

    constructor Create(AOwner: TComponent); override;
  end;

  // The following classes are specific event implementations
  // Note that T in the classes below refer to the parameter types of the execute method
  TMethodEvents = class(TEventHandlers<TNotifyEvent>)
  public
    procedure execute(Sender: TObject);
  end;

  // TMethodEvents<T>
  TMethodEvent<T> = procedure(Sender: TObject; Arg1: T) of object;
  TMethodEvents<T> = class(TEventHandlers<TMethodEvent<T>>)
  public
    procedure execute(Sender: TObject; Arg1: T);
  end;

  // Anonmous proc/func events
  TProcEvents = class(TEventHandlers<TProc>)
  public
    procedure execute;
  end;

  TProcEvents<T> = class(TEventHandlers<TProc<T>>)
  public
    procedure execute(Arg1: T);
  end;
implementation

{ TEventHandlers<T> }

procedure TEventHandlers<T>.addHandler(handler: T);
begin
  fList.Obj.Add(TObj<TEventHandler>.CreateInstance(TEventHandler.Create(nil, handler)));
end;

procedure TEventHandlers<T>.clear;
begin
  fList.Obj.Clear;
end;

constructor TEventHandlers<T>.Create(AOwner: TComponent);
begin
  inherited;
  createList;
end;

procedure TEventHandlers<T>.createList;
begin
  fList := TObj<TIEventHandlerList>.CreateInstance(TIEventHandlerList.Create);
end;

procedure TEventHandlers<T>.removeHandler(handler: T);
begin
  for var i := fList.Obj.Count-1 downto 0 do
  begin
    if (@fList.Obj[i].Obj.eventHandler = @handler) then
      fList.Obj.Delete(i);
  end;

end;

procedure TEventHandlers<T>.setHandler(handler: T);
begin
  clear;
  fList.Obj.Add(TObj<TEventHandler>.CreateInstance(TEventHandler.Create(nil, handler)));
end;

procedure TEventHandlers<T>.subscribe(subscriber: TObject; handler: T);
begin
  fList.Obj.Add(TObj<TEventHandler>.CreateInstance(TEventHandler.Create(subscriber, handler)));

  if (subscriber is TComponent) then
    TComponent(subscriber).FreeNotification(self);
end;

procedure TEventHandlers<T>.unsubscribe(subscriber: TObject);
begin
  if not assigned(subscriber) then
    exit;

  for var i := fList.Obj.Count-1 downto 0 do
  begin
    if (fList.Obj[i].Obj.subscriber = subscriber) then
      fList.Obj.Delete(i);
  end;
end;

procedure TEventHandlers<T>.unsubscribe(subscriber: TObject; handler: T);
begin
  if not(assigned(subscriber) or (@handler <> nil) ) then
    exit;

  if not assigned(subscriber) then
  begin
    removeHandler(handler);
    exit;
  end
  else if (@handler = nil) then
    unsubscribe(subscriber)

  else
  begin
    for var i := fList.Obj.Count-1 downto 0 do
      if (fList.Obj[i].Obj.subscriber = subscriber) and (@fList.Obj[i].Obj.eventHandler = @handler) then
      fList.Obj.Delete(i);
  end
end;

procedure TEventHandlers<T>.Notification(AComponent: TComponent; Operation: TOperation);
begin
  if assigned(AComponent) and (Operation = TOperation.opRemove) then
    self.unsubscribe(aComponent);

  inherited;
end;

{ TEventHandlers<T>.TEventHandler }

constructor TEventHandlers<T>.TEventHandler.Create(obj: TObject; handler: T);
begin
  subscriber:= obj;
  eventHandler := handler;
end;

{ TMethodEvents }
procedure TMethodEvents.execute(Sender: TObject);
begin
  for var eventHandler in fList.Obj do
    TThread.Queue(nil,
        procedure
        begin
          eventHandler.Obj.EventHandler(Sender);
        end
    );
end;

{ TMethodEvents<T> }

procedure TMethodEvents<T>.execute(Sender: TObject; Arg1: T);
begin
  for var eventHandler in fList.Obj do
    TThread.Queue(nil,
        procedure
        begin
          eventHandler.Obj.EventHandler(Sender, Arg1);
        end
    );
end;

{ TProcEvents }

procedure TProcEvents.execute;
begin
  for var eventHandler in fList.Obj do
    TThread.Queue(nil,
        procedure
        begin
          eventHandler.Obj.EventHandler();
        end
    );
end;

{ TProcEvents<T> }

procedure TProcEvents<T>.execute(Arg1: T);
begin
  for var eventHandler in fList.Obj do
    TThread.Queue(nil,
        procedure
        begin
          eventHandler.Obj.EventHandler(Arg1);
        end
    );
end;

end.
