unit RestMessageCenter;

interface
uses
  EventHandlers, IObj, Generics.Collections,

  Datasnap.DSCommon, Datasnap.DSClientRest, JSON, Rest.Json,
  JSON.Readers, JSON.Writers, JSON.Builders, JSON.BSON,
  System.Classes, System.SysUtils;

type
  TRestMessageCenter = class(TComponent)
  type
    TCallBackObject = class
    public
      callbackID : string;
      text : string;

      constructor create(pCallbackID, pText : string);
    end;

    TChannelContext = class
    private
      FName : string;
      FChannel : TDSRestClientChannel;
      FCallback : TDSRestClientCallback;
      FOnMessage : IObj<TProcEvents<string>>;

      procedure releaseCallbackChannel;
    public
      constructor Create(const name: string);
      destructor Destroy; override;
    end;
    IChannelContext = IObj<TChannelContext>;

  private
    FCallbackID : string;
    FConnection: TDSRestConnection;
    FContextList: IObj<TDictionary<string, IChannelContext>>;

    function CheckConnection : boolean;
    procedure SetConnection(const Value: TDSRestConnection);
    function getContext(const channelName: string): IChannelContext;
    procedure CreateContextList;

    procedure CreateCallbackChannel(context: TChannelContext);
    procedure CreateOnMessage(context: TChannelContext);
    procedure CreateCallback(context: TChannelContext);
    function GetOnMessage(channelName: string): TProcEvents<string>;
  public
    property Connection : TDSRestConnection read FConnection write SetConnection;
    procedure CreateChannel(const channelName: string);
    property OnMessage[channelName: string]: TProcEvents<string> read GetOnMessage;
    procedure broadCast(const channelName, msg : string);

    constructor Create(AOwner: TComponent); override;
  end;


implementation

{ TRestMessageCenter }

procedure TRestMessageCenter.broadCast(const channelName, msg: string);
begin
  var context := getContext(channelName);
  if not assigned(context) then
    exit;

  var cb := TObj<TCallbackObject>.CreateInstance(TCallbackObject.Create(FCallbackID, msg));
  var jObj := TObj<TJSONObject>.CreateInstance(TJSON.ObjectToJsonObject(cb.Obj));
  
  context.Obj.FChannel.Broadcast(jObj.Obj);
  context.Obj.FOnMessage.Obj.execute(msg);  // Process the message within the app . No need to wait for the broadcast message

end;

function TRestMessageCenter.CheckConnection: boolean;
begin
  result := false;
  if FConnection = nil then
    exit;

  result := (FConnection.Protocol <> '')
          and (FConnection.Host <> '')
          and (FConnection.Port <> 0);
end;

constructor TRestMessageCenter.Create(AOwner: TComponent);
begin
  inherited;
  FCallbackID := TDSSessionHelper.GenerateSessionId;
  CreateContextList;
end;

procedure TRestMessageCenter.CreateCallback(context: TChannelContext);
begin
  context.FCallback :=
            TDSRestClientCallback.Create(context.FChannel, FCallbackID,
                function(AValue: TJSONValue; ADataType: string): Boolean
                begin
                  if (AValue.FindValue('callbackID') <> nil) and (AValue.FindValue('text') <> nil) then
                  begin
                    var cb := TObj<TCallbackObject>.CreateInstance(TJSON.JsonToObject<TCallbackObject>(TJSONObject(AValue), []));
                    if Not SameText(cb.Obj.callbackID, FCallbackID) then
                      context.FOnMessage.Obj.execute(cb.Obj.text); //do something with a string representation of the message
                  end;

                  Result := true;
                end
            );

  if context.FChannel.Connected then
    context.FChannel.RegisterCallback(context.FCallback)
  else
    context.FChannel.Connect(context.fCallback);
end;

procedure TRestMessageCenter.CreateCallbackChannel(context: TChannelContext);
begin
  context.FChannel := TDSRestClientChannel.Create(FCallbackID, context.FName, FConnection);
end;

procedure TRestMessageCenter.CreateChannel(const channelName: string);
begin
  if not CheckConnection() then
    exit;

  var lChannelName := lowerCase(channelName);

  if fContextList.Obj.ContainsKey(lChannelName) then
    exit;

  var lContext := TObj<TChannelContext>.createInstance(TChannelContext.Create(channelName));
  createCallbackChannel(lContext.Obj);
  createOnMessage(lContext.Obj);
  createCallback(lContext.Obj);

  fContextList.Obj.Add(lChannelName, lContext);
end;

procedure TRestMessageCenter.CreateContextList;
begin
  FContextList := TObj<TDictionary<string, IChannelContext>>.CreateInstance(TDictionary<string, IChannelContext>.Create());
end;

procedure TRestMessageCenter.CreateOnMessage(context: TChannelContext);
begin
  context.FOnMessage := TObj<TProcEvents<string>>.CreateInstance(TProcEvents<string>.Create(nil));
end;

function TRestMessageCenter.getContext(const channelName: string): IChannelContext;
begin
  var lChannelName := lowerCase(channelName);
  result := FContextList.Obj[lChannelName];
end;

function TRestMessageCenter.GetOnMessage(channelName: string): TProcEvents<string>;
begin
  result := getContext(channelName).Obj.FOnMessage.Obj;
end;

procedure TRestMessageCenter.Setconnection(const Value: TDSRestConnection);
begin
  FConnection := Value;
end;

{ TRestMessageCenter.TChannelContext }

constructor TRestMessageCenter.TChannelContext.Create(const name: string);
begin
  fName := name;
end;

destructor TRestMessageCenter.TChannelContext.Destroy;
begin
  releaseCallbackChannel;
  inherited;
end;

procedure TRestMessageCenter.TChannelContext.releaseCallbackChannel;
begin
  FChannel.Free;
  FChannel := nil;
  fCallback := nil;
end;

{ TRestMessageCenter.TCallBackObject }

constructor TRestMessageCenter.TCallBackObject.create(pCallbackID,
  pText: string);
begin
  callbackID := pCallbackID;
  text := pText;
end;

end.
